#
# $Id: mdstandalone.py,v 1.1 2005/08/26 10:16:28 koblitz Exp $
#
import mdinterface
from mdinterface import CommandException, MDInterface
from mdparser import MDTable, MDParser
import os
import re
import time
import errno

DEBUG = False

def visitLocksRemove(arg, directory, files):
    age = arg
    for name in files:
        if name != 'LOCK':
            continue
        filename = os.path.join(directory, name)
        mtime = os.stat(filename).st_mtime
        if age > 0 and mtime > age:
            continue
        os.remove(filename)


def visitLocksList(arg, directory, files):
    (age, root, lines) = arg
    for name in files:
        if name != 'LOCK':
            continue
        filename = os.path.join(directory, name)
        mtime = os.stat(filename).st_mtime
        if age > 0 and mtime > age:
            continue               
        line = filename[len(root)-1:] # Remove path to root directory
        line = os.path.normpath(line)
        line = line[0: len(line)-5] # Remove /LOCK
        if len(line) == 0:
            line = "/"
        line = os.path.normpath(line)
        lines.append(line)

    

class MDStandalone (mdinterface.MDInterface):
    def __mkdir(self, newdir):
        """ works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
        """
        if os.path.isdir(newdir):
            pass
        elif os.path.isfile(newdir):
            raise mdinterface.CommandException(16, "Directory exists")
        try:
            os.mkdir(newdir)
        except OSError, e:
            if e[0] == errno.EPERM or e[0] == errno.EACCES:
                raise mdinterface.CommandException(4, "Could not create dir: Permission denied")
            else:
                raise mdinterface.CommandException(16, "Directory exists")


    def __init__(self, root):
        self.root=os.path.normpath(root)
        if self.root[-1] != '/':
            self.root += '/'
        self.rows=[]
        self.rowPtr=0
        self.tables = {}
        self.locks = {}
        self.currentDir = '/'
#        os.stat_float_times(True)


    def __readEntries(self, dirname):
        attributes = []
        entries = []
        name = os.path.normpath(self.root+dirname + '/ENTRIES')
        try:
            f=open(name, 'r')
        except IOError, e:
            return attributes, entries
        line = f.readline()
        if not line:
            f.close
            return attributes, entries
        line = line.rstrip('\n')
        if line :
            attributes = line.split('\0')
        i = 0
        while 1:
            line = f.readline()
            line = line.rstrip('\n')
            if not line:
                break
            fields = line.split('\0')
            entries.append(fields)
            i=i+1
        f.close()
        return attributes, entries


    def __writeEntries(self, dirname, attributes, entries):
        name = os.path.normpath(self.root + dirname + '/ENTRIES')
        if not attributes and not entries:
            if os.path.exists(name):
                os.remove(name)
            if os.path.exists(name+'~'):
                os.remove(name+'~')
            return
        f=open(name+'_', 'w')
        line=''
        for field in attributes:
            if len(line):
                line += '\0'
            line += field
        f.write(line + '\n')
        for entry in entries:
            if not len(entry):
                continue
            line=''
            for field in entry:
                if len(line):
                    line += '\0'
                line += str(field)
            f.write(line + '\n')
        f.close()
        try:
            newtime = os.stat(name+'_').st_mtime
            if os.path.exists(name):
                oldtime = os.stat(name).st_mtime
            else:
                oldtime = 0
            if newtime <= oldtime:
                os.utime(name, (oldtime+1, oldtime+1))
            os.rename(name+'_', name)
        except OSError, e:
            # Windows does not do atomic moves, need to
            print "WARNING, no atomic renames"
            os.remove(name)
            os.rename(name+'_', name)


    def __isDir(self, dirname):
        name = os.path.normpath(self.root+dirname)
        return os.path.isdir(name)


    def __createLock(self, lockfile):
        tries = 0
        while True:
            tries = tries +1
            try:
                fd = os.open(lockfile, os.O_WRONLY | os.O_CREAT | os.O_EXCL)
            except OSError, e:
                time.sleep(0.05)
                if tries > 200:
                    return False
                continue
            break
        pid = os.getpid()
        os.write(fd, str(pid))
        os.close(fd)
        return True


    def __removeLock(self, lockfile):
        os.remove(lockfile)


    def __lockTable(self, table):
        # For another idea
        # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/65203
        #fcntl.flock(tableFile.fileno(), fcntl.LOCK_EX)
        lockfile = os.path.normpath(self.root + table + '/LOCK')
        if self.__createLock(lockfile):
            self.locks[table] = True
            return True
        return False


    def __unlockTable(self, table):
        if not self.locks[table]:
            return
        lockfile = os.path.normpath(self.root + table + '/LOCK')
        self.__removeLock(lockfile)
        self.locks[table] = False



    def __getTimestamp(self, table):
        name = os.path.normpath(self.root + table + '/ENTRIES')
        if not os.path.exists(name):
            return 0
        return os.stat(name).st_mtime


    def __absolutePath(self, table):
        if DEBUG: print '__absolutePath for ', table
        if len(table) and table[0] == '/':
            if DEBUG: print '__absolutePath: returning', table
            return table
        table = os.path.normpath(self.currentDir + '/' + table)
        # Replace // -> /
        p =  table.find ('//')
        while p > -1:
            table=table[:p]+table[p+1:]
            p = table.find ('//')
        if DEBUG: print '__absolutePath: returning', table
        return table


    def __systemPath(self, path):
        path = os.path.normpath(self.root + path)
        if DEBUG: print '__systemPath returns: ', path
        return path


    def __loadTable(self, table, update = False):
        # Absolute path of table
        if DEBUG: print 'Loading table', table
        table = self.__absolutePath(table)
        
        # Check if table in cache
        if self.tables.has_key(table):
            if self.tables[table].timestamp >= self.__getTimestamp(table):
                return self.tables[table], table
        
        if DEBUG: print '... table not in chache', table
        
        # Table not in cache
        if not os.path.isdir(os.path.normpath(self.root+table)):
            raise CommandException(1, 'File or directory does not exist')
        if update:
            if not self.__lockTable(table):
                raise CommandException(9, 'Could not acquire table lock')
        else:
            self.locks[table] = False
        attributes, entries=self.__readEntries(table)
        self.tables[table] = MDTable(attributes, entries)
        return self.tables[table], table


    def __saveTable(self, table):
        mdtable = self.tables[table]
        mdtable.update()
        attributes = mdtable.attributes
        entries = mdtable.entries
        # Check for lock here!
        self.__writeEntries(table, attributes, entries)
        if self.locks[table]:
            self.__unlockTable(table)


    def releaseAllLocks(self):
        for e in self.locks:
            self.__unlockTable(e)

        
    def eot(self):
        return self.rowPtr >= len(self.rows)


    def addEntries(self, entries):
        self.tables = {}
        try:
            tablename = os.path.dirname(entries[0])
            mdtable, tablename = self.__loadTable(tablename, True)
            entryDict = {}
            for i in range(0, len(mdtable.entries)):
                entryDict[mdtable.entries[i][0]]=i
            emptyAttr = [''] * len(mdtable.attributes)
            for entry in entries:
                if entryDict.has_key(os.path.basename(entry)):
                    raise CommandException(15, "Entry exists")
                row = [os.path.basename(entry)]
                row.extend(emptyAttr)
                mdtable.entries.append(row)
            self.__saveTable(tablename)
        finally:
            self.releaseAllLocks()


    def addAttr(self, file, name, t):
        self.tables = {}
        try:
            tablename = self.__absolutePath(file)
            if not self.__isDir(file):
                tablename = os.path.dirname(file)
            mdtable, tablename = self.__loadTable(tablename, True)
            if mdtable.attributeDict.has_key(name):
                raise CommandException(15, "Attribute exists")
            mdtable.attributes.append(name+" "+t)
            mdtable.update()
            for i in range(0, len(mdtable.entries)):
                mdtable.entries[i].append('')
            self.__saveTable(tablename)
        finally:
            self.releaseAllLocks()
        

    def listAttr(self, file):
        self.tables = {}
        dirname = file
        if not self.__isDir(file):
            dirname = os.path.dirname(file)
        mdtable, dirname = self.__loadTable(dirname)
        names = []
        types = []
        for i in range(0, len(mdtable.attributes)):
            pair = mdtable.attributes[i].split(' ', 1)
            names.append(pair[0])
            types.append(pair[1])
        return names, types


    def addEntry(self, file, keys, values):
        self.tables = {}
        try:
            tablename, entry = os.path.split(file)
            mdtable, tablename = self.__loadTable(tablename, True)
            entryDict = {}
            for i in range(0, len(mdtable.entries)):
                entryDict[mdtable.entries[i][0]]=i
            if entryDict.has_key(entry):
                raise CommandException(15, "Entry exists")
            e=[''] * (len(mdtable.attributes)+1)
            e[0]=entry
            for i in range(0, len(keys)):
                e[mdtable.attributeDict[keys[i]]+1]=values[i]
            mdtable.entries.append(e)
            self.__saveTable(tablename)
        finally:
            self.releaseAllLocks()


    def getattr(self, file, attrs):
        self.tables = {}
        tablename, entry = os.path.split(file)
        mdtable, tablename = self.__loadTable(tablename)
        self.rows = []
        self.rowPtr = 0
        pattern = entry.replace('*', '.*')
        pattern = pattern.replace('?', '.')
        for i in range(0, len(mdtable.entries)):
            if re.match(pattern, mdtable.entries[i][0]):
                row = []
                row.append(mdtable.entries[i][0])
                for j in range(0, len(attrs)):
                    index = mdtable.attributeDict[attrs[j]]
                    row.append(mdtable.entries[i][index+1])
                self.rows.append(row)
        self.rowPtr = 0


    def getEntry(self):
        entry, row = self.rows[self.rowPtr][0], self.rows[self.rowPtr][1:]
        self.rowPtr = self.rowPtr + 1
        return entry, row

    
    def createDir(self, dirname):
        if DEBUG: print 'createDir ', dirname
        newdir = self.__systemPath(self.__absolutePath(dirname))
        self.__mkdir(newdir)


    def listEntries(self, files):
        if DEBUG: print 'listEntries ', files
        self.tables = {}
        files = self.__absolutePath(files)
        if self.__isDir(files):
            dirname = files
            entry = "*"
        else:
            dirname, entry = os.path.split(files)
            
        # List directories, first
        dirs = os.listdir(self.__systemPath(dirname))
        for d in dirs:
            if self.__isDir(dirname + '/' +d):
                row = []
                row.append(dirname + '/' +d)
                row.append('collection')
                self.rows.append(row)
        
        # Now list entries
        mdtable, dirname = self.__loadTable(dirname)
        pattern = entry.replace('*', '.*')
        pattern = pattern.replace('?', '.')
        for i in range(0, len(mdtable.entries)):
            if re.match(pattern, mdtable.entries[i][0]):
               row = []
               row.append(mdtable.entries[i][0])
               row.append('entry')
               self.rows.append(row)


    def removeAttr(self, file, name):
        if DEBUG: print 'removeAttr ', file, name
        self.tables = {}
        try:
            dirname = file
            if not self.__isDir(file):
                dirname = os.path.dirname(file)
            mdtable, dirname = self.__loadTable(dirname, True)
            if not mdtable.attributeDict.has_key(name):
                raise CommandException(10, "No such key")
            del mdtable.attributes[mdtable.attributeDict[name]]
            self.__saveTable(dirname)
        finally:
            self.releaseAllLocks()
        

    def rm(self, path):        
        if DEBUG: print 'rm ', path
        self.tables = {}
        try:
            dirname, entry = os.path.split(path)
            mdtable, dirname = self.__loadTable(dirname, True)
            pattern = entry.replace('*', '.*')
            pattern = pattern.replace('?', '.')
            for i in range(len(mdtable.entries)-1, -1, -1):  
                if re.match(pattern, mdtable.entries[i][0]):
                    del mdtable.entries[i]
            self.__saveTable(dirname)
        finally:
            self.releaseAllLocks()


    def sequenceCreate(self, name, directory, increment=1, start=1):
        name = self.__systemPath(self.__absolutePath(directory) + '/SEQ'+name)
        lockfile = name+".LOCK"
        if not self.__createLock(lockfile):
            raise CommandException(9, "Cannot lock")
        try:
            if os.path.exists(name):
                raise CommandException(15, "Sequence exists")
            f=open(name, 'w')
            line=str(start) + " " + str(increment) + " " + str(start)
            f.write(line + '\n')
            f.close()
        finally:
            self.__removeLock(lockfile)


    def sequenceNext(self, name):
        dirname, seq = os.path.split(name)
        name = self.__systemPath(self.__absolutePath(dirname) + '/SEQ'+seq)
        lockfile = name+".LOCK"
        if not self.__createLock(lockfile):
            raise CommandException(9, "Cannot lock")
        try:
            try:
                f=open(name, 'r')
            except IOError, e:
                raise CommandException(17, "Not a sequence")
            line = f.readline()
            if not line:
                raise CommandException(17, "Not a sequence")
            start = 0
            increment = 0
            now = 0
            start, increment, now = line.split()
            newval = int(now) + int(increment)
            f.close()
            f=open(name, 'w')
            line=str(start)+" " + str(increment) + " " + str(newval)
            f.write(line + '\n')
            f.close()
            return now
        finally:
            self.__removeLock(lockfile)

        
    def sequenceRemove(self, name):
        dirname, seq = os.path.split(name)
        name = self.__systemPath(self.__absolutePath(dirname) + '/SEQ'+seq)
        lockfile = name+".LOCK"
        if not self.__createLock(lockfile):
            raise CommandException(9, "Cannot lock")
        try:
            if not os.path.isfile(name):
                raise CommandException(17, "Not a sequence")
            os.remove(name)
        finally:
            self.__removeLock(lockfile)
        
    
    def removeDir(self, dirname):
        if DEBUG: print 'removeDir ', dirname
        name = self.__absolutePath(dirname)
        try:
            os.rmdir(self.__systemPath(name))
        except OSError, e:
            if e[0] == errno.ENOENT:
                raise mdinterface.CommandException(1, "Directory not found")
            else:
                raise mdinterface.CommandException(11, "Directory not empty")                                    
    
        
    def selectAttr(self, attrs, query):
        if DEBUG: print 'selectAttr ', attrs, query
        self.tables = {}
        self.rows = []
        self.rowPtr = 0

        tList = []
        aList = []
        for a in attrs:
            table, attr = a.split(':', 1)
            mdtable, table = self.__loadTable(table)
            tList.append(table)
            aList.append(attr)
        
        parser = MDParser(query, self.tables,
                          self.currentDir, self.__loadTable)
        # First, only load necessary tables
        parser.parseWhereClause()
        allTables = []
        iterations = 0
        for k, v in self.tables.iteritems():
            v.initRowPointer()
            allTables.append(v)
            if iterations == 0:
                iterations = len(v.entries)
            else:
                iterations = iterations * len(v.entries)

        for i in range(0, iterations):
            if parser.parseWhereClause():
                row = []
                for j in range(0, len(attrs)):
                    t = self.tables[tList[j]]
                    if aList[j] == "FILE":
                        index = 0
                    else:
                        index = t.attributeDict[aList[j]]+1
                    row.append(t.entries[t.currentRow][index])
                self.rows.append(row)
            for j in range (0, len(allTables)):
                if allTables[j].currentRow < 0:
                    continue
                allTables[j].currentRow = allTables[j].currentRow + 1
                if allTables[j].currentRow < len(allTables[j].entries):
                    break
                allTables[j].currentRow = 0   
        self.rowPtr = 0


    def getSelectAttrEntry(self):
        attributes = self.rows[self.rowPtr]
        self.rowPtr = self.rowPtr + 1
        return attributes


    def splitUpdateClause(self, clause):
        #skip leading white space
        i = 0
        while i<len(clause) and (clause[i] == ' ' or clause[i] == '\t'):
            i = i +1
        clause=clause[i:]
       
        espcaped = False
        quoted = False
        i = 0
        while i < len(clause):
            if clause[i] == "'" and not escaped:
                quoted = not quoted
            if clause[i] == '/':
                escaped = not escaped
            if (clause[i] == ' ' or clause[i] == '/t') and not quoted:
                break
            i = i + 1
        if i==0 or i>=len(clause)-1:
            raise mdinterface.CommandException(3, "Invalid update statement") 
                            
        var = clause[0:i]
        exp = clause[i+1:]
        i1 = exp.find("'")    
        i2 = exp.rfind("'")   
        if i1==0 and i2 == len(exp)-1:
            exp = exp[i1+1:i2]
        i1 = var.find("'")    
        i2 = var.rfind("'")           
        if i1==0 and i2 == len(var)-1:          
            var = var[i1+1:i2]
        return var, exp


    def updateAttr(self, pattern, updateExpr, condition):
        if DEBUG: print 'updateAttr ', pattern, updateExpr, condition
        self.tables = {}
        try:
            tablename, entry = os.path.split(pattern)
            mdtable, tablename = self.__loadTable(tablename, True)
            pattern = entry.replace('*', '.*')
            pattern = pattern.replace('?', '.')

            parser = MDParser(condition, self.tables,
                              tablename, self.__loadTable)

            # First, only load necessary tables
            parser.parseWhereClause()

            # Prepare update expressions
            expressions = []
            for e in updateExpr:
                var, exp = self.splitUpdateClause(e)
                index = mdtable.attributeDict[var]+1
                p = MDParser(exp, self.tables, tablename,
                             self.__loadTable)
                parser.parseWhereClause()
                expressions.append( (index, exp, p) )

            # Find all tables which have entries and list them in allTables,
            # and count the number of iterations we need to do
            allTables = []
            iterations = 0
            for k, v in self.tables.iteritems():
                v.initRowPointer()
                if len(v.entries) == 0:
                  continue
                allTables.append(v)
                if iterations == 0:
                    iterations = len(v.entries)
                else:
                    iterations = iterations * len(v.entries)

            # Iterate over all combinations of table entries, check whether
            # update clause if fulfilled and do the update
            for i in range(0, iterations):
                if re.match(pattern, mdtable.entries[mdtable.currentRow][0]) \
                       and parser.parseWhereClause():
                    for j in range (0, len(expressions)):
                        index, exp, p = expressions[j]
                        value = p.parseStatement()
                        mdtable.entries[mdtable.currentRow][index] = value
                for j in range (0, len(allTables)):
                    allTables[j].currentRow = allTables[j].currentRow + 1
                    if allTables[j].currentRow < len(allTables[j].entries):
                        break
                    else:
                        allTables[j].currentRow = 0
            self.__saveTable(tablename)
        finally:
            self.releaseAllLocks()


    def setAttr(self, file, keys, values):
        self.tables = {}
        tablename, entry = os.path.split(file)
        mdtable, tablename = self.__loadTable(tablename, True)
        try:
            pattern = entry.replace('*', '.*')
            pattern = pattern.replace('?', '.')
            for i in range(0, len(mdtable.entries)):
                if re.match(pattern, mdtable.entries[i][0]):
                    for j in range(0, len(keys)):
                        index = mdtable.attributeDict[keys[j]]
                        mdtable.entries[i][index+1]=values[j]
                self.__saveTable(tablename)
        finally:
            self.releaseAllLocks()


    def pwd(self):
        return self.currentDir


    def cd(self, dir):
        dir = self.__absolutePath(dir)
        if not self.__isDir(dir):
            raise CommandException(1, "Not a directory")
        self.currentDir = dir


    # Removes all locks, even those of other processes recursively
    # starting at the root directory by deleting the LOCK files
    # If age is given it restricts the deletion operation on most
    # operating systems to locks older than age seconds
    def removeAllLocks(self, age = -1):
        if age > 0:
            age = time.time()-age
        os.path.walk(self.root, visitLocksRemove, age);


    # Returns a list witha all currently held locks (locks older than age if age>0)
    def listAllLocks(self, age = -1):
        if age > 0:
            age = time.time()-age
        
        lines = []
        os.path.walk(self.root, visitLocksList, (age, self.root, lines));
        return lines
